from __future__ import annotations

import enum
import time
from typing import Optional, Dict, Generic, Tuple, TypeVar, Collection, List, Union, Type, TYPE_CHECKING

from ravendb.primitives import constants
from ravendb.exceptions import exceptions
from ravendb.json.metadata_as_dictionary import MetadataAsDictionary
import OpenSSL.crypto

try:
    from collections.abc import Iterable, Sequence
except ImportError:
    from collections import Iterable, Sequence

from ravendb.tools.projection import create_entity_with_mapper
from datetime import datetime, timedelta
from enum import Enum
from threading import Timer
from copy import deepcopy
import urllib
import inspect
import json
import sys
import re

_T = TypeVar("_T")
_TKey = TypeVar("_TKey")
_TVal = TypeVar("_TVal")

_default_wildcards = {
    "-",
    "&",
    "|",
    "!",
    "(",
    ")",
    "{",
    "}",
    "[",
    "]",
    "^",
    '"',
    "'",
    "~",
    ":",
    "\\",
}
if TYPE_CHECKING:
    from ravendb.documents.conventions import DocumentConventions


class TimeUnit(Enum):
    NANOSECONDS = 1
    MICROSECONDS = 1000 * NANOSECONDS
    MILLISECONDS = 1000 * MICROSECONDS
    SECONDS = 1000 * MILLISECONDS
    MINUTES = 60 * SECONDS
    HOURS = 60 * MINUTES
    DAYS = 24 * HOURS

    def convert(self, source_duration: int, source_unit: TimeUnit):
        if self == TimeUnit.NANOSECONDS:
            return source_unit.to_nanos(source_duration)
        elif self == TimeUnit.MICROSECONDS:
            return source_unit.to_micros(source_duration)
        elif self == TimeUnit.MILLISECONDS:
            return source_unit.to_millis(source_duration)
        elif self == TimeUnit.SECONDS:
            return source_unit.to_seconds(source_duration)
        raise NotImplementedError("Unsupported TimeUnit")

    def to_nanos(self, duration: int) -> int:
        return duration * self.value

    def to_micros(self, duration: int) -> int:
        return duration * self.value / 1000

    def to_millis(self, duration: int) -> int:
        return duration * self.value / (1000 * 1000)

    def to_seconds(self, duration: int) -> int:
        return duration * self.value / (1000 * 1000 * 1000)

    def to_minutes(self, duration: int) -> int:
        return duration * self.value / (1000 * 1000 * 1000 * 60)

    def to_hours(self, duration: int) -> int:
        return duration * self.value / (1000 * 1000 * 1000 * 60 * 60)

    def to_days(self, duration: int) -> int:
        return duration * self.value / (1000 * 1000 * 1000 * 60 * 60 * 24)


class Stopwatch:
    def __init__(self):
        self.__is_running = False
        self.__elapsed_nanos = 0
        self.__start_tick = None

    @staticmethod
    def create_started() -> Stopwatch:
        return Stopwatch().start()

    @property
    def is_running(self) -> bool:
        return self.__is_running

    def start(self) -> Stopwatch:
        if self.__is_running:
            raise RuntimeError("This stopwatch is already running.")
        self.__is_running = True
        self.__start_tick = time.perf_counter_ns()
        return self

    def stop(self) -> Stopwatch:
        tick = time.perf_counter_ns()
        if not self.__is_running:
            raise RuntimeError("This stopwatch is already stopped.")
        self.__is_running = False
        self.__elapsed_nanos = (
            (tick - self.__start_tick)
            if self.__elapsed_nanos is None
            else self.__elapsed_nanos + (tick - self.__start_tick)
        )
        return self

    def reset(self) -> Stopwatch:
        self.__elapsed_nanos = 0
        self.__is_running = False
        return self

    def __take_elapsed_nanos(self) -> float:
        return (
            time.perf_counter_ns() - self.__start_tick + self.__elapsed_nanos
            if self.__is_running
            else self.__elapsed_nanos
        )

    def elapsed_micros(self) -> float:
        return self.__take_elapsed_nanos() / 1000

    def elapsed(self, desired_unit: Optional[TimeUnit] = None) -> Union[timedelta, int]:
        if desired_unit is not None:
            return desired_unit.convert(self.__elapsed_nanos, TimeUnit.NANOSECONDS)
        return timedelta(microseconds=self.elapsed_micros())


class Size:
    def __init__(self, size_in_bytes: int = None, human_size: str = None):
        self.size_in_bytes = size_in_bytes
        self.human_size = human_size

    @classmethod
    def from_json(cls, json_dict: Dict) -> Size:
        return cls(json_dict["SizeInBytes"], json_dict["HumaneSize"])


# todo: https://issues.hibernatingrhinos.com/issue/RDBC-686
class CaseInsensitiveDict(dict, Generic[_TKey, _TVal]):
    @classmethod
    def _lower_if_str(cls, key):
        return key.lower() if isinstance(key, str) else key

    def __init__(self, *args, **kwargs):
        super(CaseInsensitiveDict, self).__init__(*args, **kwargs)
        self._original_values: Dict[_TKey, _TKey] = {}
        self._convert_keys()

    def _convert_key_back(self, lowered_key: _TKey) -> _TKey:
        return self._original_values[lowered_key]

    def _convert_item_back(self, item: Tuple[_TKey, _TVal]) -> Tuple[_TKey, _TVal]:
        return self._convert_key_back(item[0]), item[1]

    def __getitem__(self, key) -> _TVal:
        return super(CaseInsensitiveDict, self).__getitem__(self.__class__._lower_if_str(key))

    def __setitem__(self, key, value):
        adjusted_key = self.__class__._lower_if_str(key)
        self._original_values[adjusted_key] = key
        super(CaseInsensitiveDict, self).__setitem__(adjusted_key, value)

    def __delitem__(self, key):
        return super(CaseInsensitiveDict, self).__delitem__(self.__class__._lower_if_str(key))

    def __contains__(self, key):
        return super(CaseInsensitiveDict, self).__contains__(self.__class__._lower_if_str(key))

    def pop(self, key, *args, **kwargs) -> _TVal:
        return super(CaseInsensitiveDict, self).pop(self.__class__._lower_if_str(key), *args, **kwargs)

    def remove(self, key, *args, **kwargs) -> Optional[_TVal]:
        return self.pop(key, *args, **kwargs) if key in self else None

    def get(self, key, *args, **kwargs) -> _TVal:
        return super(CaseInsensitiveDict, self).get(self.__class__._lower_if_str(key), *args, **kwargs)

    def setdefault(self, key, *args, **kwargs):
        return super(CaseInsensitiveDict, self).setdefault(self.__class__._lower_if_str(key), *args, **kwargs)

    def update(self, e=None, **f):
        super(CaseInsensitiveDict, self).update(self.__class__(e))
        super(CaseInsensitiveDict, self).update(self.__class__(**f))

    def _convert_keys(self):
        for k in list(self):
            v = super(CaseInsensitiveDict, self).pop(k)
            self.__setitem__(k, v)

    def keys(self) -> List[_TKey]:
        return list(self._original_values.values())


class CaseInsensitiveSet(set):
    @classmethod
    def _v(cls, value):
        return value.lower() if isinstance(value, str) else value

    def __init__(self, *args, **kwargs):
        super(CaseInsensitiveSet, self).__init__(*args, **kwargs)
        self._convert_values()

    def __contains__(self, value):
        return super(CaseInsensitiveSet, self).__contains__(self.__class__._v(value))

    def add(self, element) -> None:
        super().add(self._v(element))

    def discard(self, element) -> None:
        super().discard(self._v(element))

    def remove(self, element) -> None:
        super().remove(self._v(element))

    def update(self, e=None, **f):
        super(CaseInsensitiveSet, self).update(self.__class__(e))
        super(CaseInsensitiveSet, self).update(self.__class__(**f))

    def _convert_values(self):
        for v in self:
            super().discard(v)
            self.add(v)


class DynamicStructure(object):
    def __init__(self, **entries):
        self.__dict__.update(entries)

    def __str__(self):
        return str(self.__dict__)


class QueryFieldUtil:
    @staticmethod
    def _should_escape(s: str, is_path: bool) -> bool:
        escape = False
        inside_escaped = False

        first = True
        for c in s:
            if c in ["'", '"']:
                inside_escaped = not inside_escaped
                continue
            if first:
                first = False
                if not c.isalpha() and c not in ["_", "@"] and not inside_escaped:
                    escape = True
                    break
            else:
                if not c.isalnum() and c not in ["_", "-", "@", ".", "[", "]"] and not inside_escaped:
                    escape = True
                    break

                if is_path and c == "." and not inside_escaped:
                    escape = True
                    break

        escape |= inside_escaped
        return escape

    @staticmethod
    def escape_if_necessary(name: str, is_path: bool = False) -> str:
        if (
            not name
            or name.isspace()
            or name
            in [
                constants.Documents.Indexing.Fields.DOCUMENT_ID_FIELD_NAME,
                constants.Documents.Indexing.Fields.REDUCE_KEY_HASH_FIELD_NAME,
                constants.Documents.Indexing.Fields.REDUCE_KEY_KEY_VALUE_FIELD_NAME,
                constants.Documents.Indexing.Fields.VALUE_FIELD_NAME,
                constants.Documents.Indexing.Fields.SPATIAL_SHAPE_FIELD_NAME,
            ]
        ):
            return name

        if not QueryFieldUtil._should_escape(name, is_path):
            return name

        sb = [c for c in name]
        need_end_quote = False
        last_term_start = 0

        for i in range(len(sb)):
            c = sb[i]
            if i == 0 and not c.isalpha() and c not in ["_", "@"]:
                sb.insert(last_term_start, "'")
                need_end_quote = True
                continue

            if is_path and c == ".":
                if need_end_quote:
                    need_end_quote = False
                    sb.insert(i, "'")
                    i += 1

                last_term_start = i + 1
                continue

            if c.isalnum() and c not in ["_", "-", "@", ".", "[", "]"] and not need_end_quote:
                sb.insert(last_term_start, "'")
                need_end_quote = True
                continue

        if need_end_quote:
            sb.append("'")

        return "".join(sb)


class Utils(object):
    primitives = (int, float, bool, str, bytes, bytearray)
    mutable_collections = (list, set)
    collections_no_str = (list, set, tuple)
    primitives_and_collections = (int, float, bool, str, bytes, bytearray, list, set, tuple)

    @staticmethod
    def check_if_collection_but_not_str(instance) -> bool:
        return isinstance(instance, (Sequence, set)) and not isinstance(instance, (str, bytes, bytearray))

    @staticmethod
    def unpack_collection(items: Collection) -> List:
        results = []

        for item in items:
            if Utils.check_if_collection_but_not_str(item):
                results.extend(Utils.__unpack_collection(item))
                continue
            results.append(item)

        return results

    @staticmethod
    def quote_key(key, reserved_slash=False, reserved_at=False) -> str:
        reserved = ""
        if reserved_slash:
            reserved += "/"
        if reserved_at:
            reserved += "@"
        if key:
            return urllib.parse.quote(key, safe=reserved)
        else:
            return ""

    @staticmethod
    def unpack_iterable(iterable):
        for item in iterable:
            if isinstance(item, Iterable) and not isinstance(item, str):
                for nested in Utils.unpack_iterable(item):
                    yield nested
            else:
                yield item

    @staticmethod
    def convert_to_snake_case(name):
        s1 = re.sub("(.)([A-Z][a-z]+)", r"\1_\2", name)
        return re.sub("([a-z0-9])([A-Z])", r"\1_\2", s1).lower()

    @staticmethod
    def database_name_validation(name):
        if name is None:
            raise ValueError("None name is not valid")
        result = re.match(r"([A-Za-z0-9_\-\.]+)", name, re.IGNORECASE)
        if not result:
            raise exceptions.InvalidOperationException(
                'Database name can only contain only A-Z, a-z, "_", "." or "-" but was: ' + name
            )

    @staticmethod
    def first_or_default(iterator, func, default):
        for item in iterator:
            if func(item):
                return item
        return default

    @staticmethod
    def get_change_vector_from_header(response):
        header = response.headers.get("ETag", None)
        if header is not None and header[0] == '"':
            return header[1 : len(header) - 2]

    @staticmethod
    def import_class(name) -> Optional[Type]:
        components = name.split(".")
        module_name = ".".join(name.split(".")[:-1])
        mod = None
        try:
            mod = getattr(__import__(module_name, fromlist=[components[-1]]), components[-1])
        except (ImportError, ValueError, AttributeError):
            pass
        return mod

    @staticmethod
    def is_inherit(parent, child):
        if child is None or parent is None:
            return False
        if parent == child:
            return True
        if parent != child:
            return Utils.is_inherit(parent, child.__base__)

    @staticmethod
    def fill_with_nested_object_types(entity: object, nested_object_types: Dict[str, type]) -> object:
        for key in nested_object_types:
            attr = getattr(entity, key)
            if attr:
                try:
                    if isinstance(attr, list):
                        nested_list = []
                        for attribute in attr:
                            nested_list.append(Utils.initialize_object(attribute, nested_object_types[key]))
                        setattr(entity, key, nested_list)
                    elif nested_object_types[key] is datetime:
                        setattr(entity, key, Utils.string_to_datetime(attr))
                    elif nested_object_types[key] is timedelta:
                        setattr(entity, key, Utils.string_to_timedelta(attr))
                    else:
                        setattr(
                            entity,
                            key,
                            Utils.initialize_object(attr, nested_object_types[key]),
                        )
                except TypeError as e:
                    print(e)
                    pass
        return entity

    @staticmethod
    def initialize_object(obj: dict, object_type: Type[_T], convert_to_snake_case: bool = None) -> _T:
        initialize_dict, set_needed = Utils.make_initialize_dict(obj, object_type.__init__, convert_to_snake_case)
        try:
            o = object_type(**initialize_dict)
        except Exception as e:
            if "Id" not in initialize_dict:
                initialize_dict["Id"] = None
                o = object_type(**initialize_dict)
            else:
                raise TypeError(
                    f"Couldn't initialize object of type '{object_type.__name__}' using dict '{initialize_dict}'"
                ) from e
        if set_needed:
            for key, value in obj.items():
                setattr(o, key, value)
        return o

    @staticmethod
    def get_field_names(object_type: Type[_T]) -> List[str]:
        obj = Utils.initialize_object({}, object_type)
        return list(obj.__dict__.keys())

    @staticmethod
    def try_get_new_instance(object_type: Type[_T]) -> _T:
        try:
            return Utils.initialize_object({}, object_type)
        except Exception as e:
            raise RuntimeError(
                f"Couldn't initialize an object of the '{object_type.__name__}' class. "
                f"Using 'None' values as arguments in the __init__ method. {e.args[0]}",
                e,
            )

    @staticmethod
    def convert_json_dict_to_object(
        json_dict: dict, object_type: Optional[Type[_T]] = None, nested_object_types: Optional[Dict[str, type]] = None
    ) -> Union[DynamicStructure, _T]:
        if object_type == dict:
            return json_dict

        if object_type is None:
            return DynamicStructure(**json_dict)

        if nested_object_types is None:
            return Utils.initialize_object(json_dict, object_type, True)

        entity = DynamicStructure(**json_dict)
        entity.__class__ = object_type
        entity = Utils.initialize_object(json_dict, object_type, True)
        if nested_object_types:
            Utils.fill_with_nested_object_types(entity, nested_object_types)
        Utils.deep_convert_to_snake_case(entity)
        return entity

    @staticmethod
    def deep_convert_to_snake_case(entity):
        if entity is None:
            return
        if type(entity) in [int, float, bool, str, set, list, tuple, dict]:
            return
        changes = {}
        for key, value in entity.__dict__.items():
            # todo: i dont' like the way we try to convert to snake case content of the object dict
            new_key = Utils.convert_to_snake_case(key)
            if key != new_key:
                changes.update({(key, new_key): value})  # collect the keys that changed (snake case conversion)
            Utils.deep_convert_to_snake_case(value)
        for keys, value in changes.items():  # erase
            entity.__dict__[keys[1]] = value
            del entity.__dict__[keys[0]]

    @staticmethod
    def get_object_fields(instance: object) -> Dict[str, object]:
        return {
            name: value
            for name, value in inspect.getmembers(instance, lambda attr: not callable(attr))
            if not name.startswith("__")
        }

    @staticmethod
    def get_class_fields(object_type: type) -> Dict[str, object]:
        try:
            instance = Utils.try_get_new_instance(object_type)
        except Exception as e:
            raise RuntimeError(
                f"Unable to retrieve information about class fields. "
                f"Couldn't get instance of '{object_type.__qualname__}'.",
                e,
            )
        return Utils.get_object_fields(instance)

    @staticmethod
    def convert_to_entity(
        document: dict,
        object_type: Optional[Type[_T]],
        conventions: "DocumentConventions",
        events,
        nested_object_types=None,
    ) -> Union[_T, DynamicStructure]:
        if document is None:
            return None
        metadata = document.get("@metadata")
        original_document = deepcopy(document)
        type_from_metadata = conventions.try_get_type_from_metadata(metadata)
        mapper = conventions.mappers.get(object_type, None)

        events.before_conversion_to_entity(document, metadata, type_from_metadata)

        if object_type == dict:
            events.after_conversion_to_entity(document, document, metadata)
            return document, metadata, original_document

        if type_from_metadata is None:
            if object_type is not None:
                metadata["Raven-Python-Type"] = "{0}.{1}".format(object_type.__module__, object_type.__name__)
            else:  # no type defined on document or during load, return a dict
                dyn = DynamicStructure(**document)
                events.after_conversion_to_entity(dyn, document, metadata)
                return dyn, metadata, original_document
        else:
            object_from_metadata = Utils.import_class(type_from_metadata)
            if object_from_metadata is not None:
                if object_type is None:
                    object_type = object_from_metadata

                elif Utils.is_inherit(object_type, object_from_metadata):
                    mapper = conventions.mappers.get(object_from_metadata, None) or mapper
                    object_type = object_from_metadata
                elif object_type is not object_from_metadata:
                    # todo: Try to parse if we use projection
                    raise exceptions.InvalidOperationException(
                        f"Cannot covert document from type {object_from_metadata} to {object_type}"
                    )

        if nested_object_types is None and mapper:
            entity = create_entity_with_mapper(document, mapper, object_type)
        else:
            entity = DynamicStructure(**document)
            entity.__class__ = object_type

            entity = Utils.initialize_object(document, object_type)

            if nested_object_types:
                Utils.fill_with_nested_object_types(entity, nested_object_types)

        if "Id" in entity.__dict__:
            entity.Id = metadata.get("@id", None)
        events.after_conversion_to_entity(entity, document, metadata)
        return entity, metadata, original_document

    @staticmethod
    def make_initialize_dict(document, entity_init, convert_to_snake_case=None):
        """
        This method will create an entity from document
        In case convert_to_snake_case will convert document keys to snake_case
        convert_to_snake_case can be dictionary with special words you can change ex. From -> from_date
        """
        if convert_to_snake_case:
            convert_to_snake_case = {} if convert_to_snake_case is True else convert_to_snake_case
            try:
                converted_document = {}
                for key in document:
                    converted_key = convert_to_snake_case.get(key, key)
                    converted_document[converted_key if key == "Id" else Utils.convert_to_snake_case(converted_key)] = (
                        document[key]
                    )
                document = converted_document
            except:
                pass

        if entity_init is None:
            return document

        set_needed = False
        entity_initialize_dict = {}
        args, __, keywords, defaults, _, _, _ = inspect.getfullargspec(entity_init)
        if (len(args) - 1) > len(document):
            remainder = len(args)
            if defaults:
                remainder -= len(defaults)
            for i in range(1, remainder):
                entity_initialize_dict[args[i]] = document.get(args[i], None)
            for i in range(remainder, len(args)):
                entity_initialize_dict[args[i]] = document.get(args[i], defaults[i - remainder])
        else:
            if keywords:
                entity_initialize_dict = document
            else:
                for key in document:
                    if key in args:
                        entity_initialize_dict[key] = document[key]
            if not entity_initialize_dict and len(args) - 1 > 0:
                set_needed = True
                for key in args[1:]:
                    entity_initialize_dict[key] = None
        return entity_initialize_dict, set_needed

    @staticmethod
    def dict_to_bytes(the_dict):
        json_dict = json.dumps(the_dict)
        return bytes(json_dict, encoding="utf-8")

    @staticmethod
    def dict_to_string(dictionary):
        builder = []
        for item in dictionary:
            if sys.version_info.major > 2 and isinstance(dictionary[item], bytes):
                dictionary[item] = dictionary[item].decode("utf-8")
            builder.append("{0}={1}".format(item, dictionary[item]))
        return ",".join(item for item in builder)

    @staticmethod
    def datetime_to_string(datetime_obj: datetime, return_none_if_none: bool = True):
        add_suffix = "0" if datetime_obj != datetime.max else "9"
        return (
            datetime_obj.strftime(f"%Y-%m-%dT%H:%M:%S.%f{add_suffix}")
            if datetime_obj
            else None if return_none_if_none else ""
        )

    @staticmethod
    def start_a_timer(interval, function, args=None, name=None, daemon=False):
        timer = Timer(interval, function, args)
        timer.daemon = daemon
        if name is not None:
            timer.name = name
        timer.start()

        return timer

    @staticmethod
    def string_to_datetime(datetime_str):
        if datetime_str is None:
            return None
        try:
            if datetime_str.endswith("Z"):
                datetime_str = datetime_str[:-1]
            datetime_s = datetime.strptime(datetime_str, "%Y-%m-%dT%H:%M:%S.%f")
        except ValueError:
            datetime_s = datetime.strptime(datetime_str[:-1], "%Y-%m-%dT%H:%M:%S.%f")

        return datetime_s

    @staticmethod
    def timedelta_tick(td):
        return int(td.total_seconds() * 10000000)

    @staticmethod
    def string_to_timedelta(timedelta_str):
        pattern = r"(?:(-?\d+)[.])?(\d{2}):(\d{2}):(\d{2})(?:.(\d+))?"
        timedelta_initialize = None
        m = re.match(pattern, timedelta_str, re.IGNORECASE)
        if m:
            timedelta_initialize = {
                "days": 0 if m.group(1) is None else int(m.group(1)),
                "hours": 0 if m.group(2) is None else int(m.group(2)),
                "minutes": 0 if m.group(3) is None else int(m.group(3)),
                "seconds": 0 if m.group(4) is None else int(m.group(4)),
                "microseconds": 0 if m.group(5) is None else int(m.group(5)),
            }
        if timedelta_initialize:
            return timedelta(**timedelta_initialize)
        return None

    @staticmethod
    def timedelta_to_str(timedelta_obj: timedelta):
        timedelta_str = None
        if isinstance(timedelta_obj, timedelta):
            timedelta_str = ""
            total_seconds = timedelta_obj.seconds
            days = timedelta_obj.days
            hours = total_seconds // 3600
            minutes = (total_seconds // 60) % 60
            seconds = (total_seconds % 3600) % 60
            microseconds = timedelta_obj.microseconds
            if days > 0:
                timedelta_str += "{0}.".format(days)
            timedelta_str += "{:02}:{:02}:{:02}".format(hours, minutes, seconds)
            if microseconds > 0:
                timedelta_str += f".{str(microseconds).rjust(6, '0')}"
        return timedelta_str

    @staticmethod
    def escape(term: str, allow_wild_cards: bool = False, make_phrase: bool = False) -> str:
        return Utils.__escape_internal(term, _default_wildcards if allow_wild_cards else None, make_phrase)

    @staticmethod
    def escape_skip(term: str, skipped_wild_cards: List[str], make_phrase: bool = False) -> str:
        return Utils.__escape_internal(term, _default_wildcards.difference(skipped_wild_cards), make_phrase)

    @staticmethod
    def __escape_internal(term: str, wild_cards: Collection[str] = None, make_phrase: bool = False) -> str:
        allow_wild_cards = wild_cards is None
        if wild_cards is None:
            wild_cards = []

        if not term:
            return '""'
        start = 0
        length = len(term)
        buffer = ""
        if length >= 2 and term[0] == "/" and term[1] == "/":
            buffer += "//"
            start = 2
        i = start
        while i < length:
            ch = term[i]
            if ch == "*" or ch == "?":
                if allow_wild_cards:
                    i += 1
                    continue

            if ch in wild_cards:
                if i > start:
                    buffer += term[start : i - start]

                buffer += "\\{0}".format(ch)
                start = i + 1

            elif ch == " " or ch == "\t":
                if make_phrase:
                    return f'"{Utils.escape(term, allow_wild_cards, False)}"'

            i += 1
        if length > start:
            buffer += term[start : length - start]

        return buffer

    @staticmethod
    def escape_collection_name(collection_name: str):
        special = ["'", '"', "\\"]
        position = 0
        buffer = []
        for char in collection_name:
            if char in special:
                buffer.append("\\")
            buffer.append(char)
            position += 1

        return "".join(buffer)

    @staticmethod
    def pfx_to_pem(pem_path, pfx_path, pfx_password):
        """
        @param pem_path: Where to create the pem file
        @param pfx_path: The path to the pfx file
        @param pfx_password: The password to pfx file
        """
        with open(pem_path, "wb") as pem_file:
            with open(pfx_path, "rb") as pfx_file:
                pfx = pfx_file.read()
            p12 = OpenSSL.crypto.load_pkcs12(pfx, pfx_password)
            pem_file.write(OpenSSL.crypto.dump_privatekey(OpenSSL.crypto.FILETYPE_PEM, p12.get_privatekey()))
            pem_file.write(OpenSSL.crypto.dump_certificate(OpenSSL.crypto.FILETYPE_PEM, p12.get_certificate()))
            ca = p12.get_ca_certificates()
            if ca is not None:
                for cert in ca:
                    # In python 3.6* we need to save the ca to ?\lib\site-packages\certifi\cacert.pem.
                    pem_file.write(OpenSSL.crypto.dump_certificate(OpenSSL.crypto.FILETYPE_PEM, cert))
        return pem_path

    @staticmethod
    def get_cert_file_fingerprint(pem_path):
        with open(pem_path, "rb") as pem_file:
            cert = OpenSSL.crypto.load_certificate(OpenSSL.crypto.FILETYPE_PEM, pem_file.read())
            return str(cert.digest("sha1"))

    @staticmethod
    def pem_to_crt_and_key(pem_path: str) -> Tuple[str, str]:
        with open(pem_path, "rb") as file:
            content = file.read()
            if "BEGIN CERTIFICATE" not in content:
                raise ValueError(
                    f"Invalid file. File stored under the path '{pem_path}' isn't valid .pem certificate. "
                    f"BEGIN CERTIFICATE header wasn't found."
                )
            if "BEGIN RSA PRIVATE KEY" not in content:
                raise ValueError(
                    f"Invalid file. File stored under the path '{pem_path}' isn't valid .pem certificate. "
                    f"BEGIN RSA PRIVATE KEY header wasn't found."
                )

            content = content.decode("utf-8")
            crt = content.split("-----BEGIN RSA PRIVATE KEY-----")[0]
            key = content.split("-----END CERTIFICATE-----")[1]

        return crt, key

    @staticmethod
    def index_of_any(text, any_of):
        """
        @param str text: The text we want to check
        @param list any_of: list of char
        :returns False if text nit
        """
        result = -1
        if not text or not any_of:
            return result

        any_of = list(set(any_of))
        i = 0
        while i < len(text) and result == -1:
            for c in any_of:
                if c == text[i]:
                    result = i
                    break
            i += 1
        return result

    @staticmethod
    def contains_any(text, any_of):
        """
        @param text: The text we want to check
        :type str
        @param any_of: list of char
        :type list
        :returns False if text nit
        """
        if not text or not any_of:
            return False

        any_of = list(set(any_of))
        for c in any_of:
            if c in text:
                return True
        return False

    @staticmethod
    def sort_iterable(iterable, key=None):
        """
        This function will take an iterable and try to sort it by the key
        if the key is not given will use the sorted default one.
        return a generator
        """
        if not iterable:
            return iterable
        yield sorted(iterable, key=key)

    @staticmethod
    def json_default(o):
        if o is None:
            return None

        if isinstance(o, datetime):
            return Utils.datetime_to_string(o)
        elif isinstance(o, timedelta):
            return Utils.timedelta_to_str(o)
        elif isinstance(o, Enum):
            return o.value
        elif isinstance(o, MetadataAsDictionary):
            return o.metadata
        elif getattr(o, "__dict__", None):
            return o.__dict__
        elif isinstance(o, set):
            return list(o)
        elif isinstance(o, int) or isinstance(o, float):
            return str(o)

        else:
            raise TypeError(repr(o) + " is not JSON serializable")

    @staticmethod
    def get_default_value(object_type: Type[_T]) -> _T:
        if object_type == bool:
            return False
        elif object_type == str:
            return ""
        elif object_type == bytes:
            return bytes(0)
        elif object_type == int:
            return int(0)
        elif object_type == float:
            return float(0)
        return None

    @staticmethod
    def dictionarize(obj: object) -> dict:
        dictionarized = {"__name__": obj.__class__.__name__}
        dictionarized.update(obj.__dict__)
        to_update = {}
        for k, v in dictionarized.items():
            if v is not None and not isinstance(
                v, (bool, float, str, int, bytes, bytearray, list, set, dict, enum.Enum)
            ):
                if "__str__" in v.__dict__:
                    to_update.update({k: str(v)})
                else:
                    to_update.update({k: Utils.dictionarize(v)})
        dictionarized.update(to_update)
        return dictionarized

    @staticmethod
    def entity_to_dict(entity, default_method) -> dict:
        return json.loads(json.dumps(entity, default=default_method))

    @staticmethod
    def add_hours(date: datetime, hours: int):
        return date + timedelta(hours=hours)

    @staticmethod
    def add_days(date: datetime, days: int):
        return date + timedelta(days=days)

    @staticmethod
    def add_minutes(date: datetime, minutes: int):
        return date + timedelta(minutes=minutes)

    @staticmethod
    def get_unix_time_in_ms(date: datetime) -> int:
        return int(date.timestamp() * 1000)



from typing import List

from ravendb.tools.utils import Utils


def to_web_socket_path(path: str) -> str:
    return path.replace("http://", "ws://").replace("https://", "wss://")


def is_identifier(token: str, start: int = 0, length=None) -> bool:
    if length == 0 or length > 256:
        return False
    if token[start].isalpha() and token[start] != "_":
        return False

    for i in range(1, length):
        if not token[start + i].isalnum() and token[start + i] != "_":
            return False

    return True


def escape_string(builder: List[str], value: str) -> None:
    if not value or value.isspace():
        return

    _escape_string_internal(builder, value)


def _escape_string_internal(builder: List[str], value: str) -> None:
    escaped = Utils.escape(value)  # todo: escape_ecma_script
    builder.append(escaped)



from typing import List, Any, Dict


class BatchCommandResult:
    def __init__(self, results, transaction_index):
        self.results: [None, list] = results
        self.transaction_index: [None, int] = transaction_index


class JsonArrayResult:
    def __init__(self, results: List = None):
        self.results = results

    @classmethod
    def from_json(cls, json_dict: Dict[str, Any]):
        return cls(json_dict["Results"])



from __future__ import annotations

import datetime
import enum
import json
from abc import abstractmethod
from typing import Generic, TypeVar, TYPE_CHECKING, Optional, List, Dict, Any
import requests

from ravendb.primitives import constants
from ravendb.documents.operations.operation import Operation
from ravendb.serverwide.database_record import DatabaseRecordWithEtag, DatabaseRecord
from ravendb.serverwide.misc import DatabaseTopology
from ravendb.tools.utils import Utils
from ravendb.http.raven_command import RavenCommand, VoidRavenCommand
from ravendb.util.util import RaftIdGenerator
from ravendb.http.topology import RaftCommand


if TYPE_CHECKING:
    from ravendb.http.server_node import ServerNode
    from ravendb.http.request_executor import RequestExecutor
    from ravendb.documents.conventions import DocumentConventions

T = TypeVar("T")


class DatabasePromotionStatus(enum.Enum):
    WAITING_FOR_FIRST_PROMOTION = "WAITING_FOR_FIRST_PROMOTION"
    NOT_RESPONDING = "NOT_RESPONDING"
    INDEX_NOT_UP_TO_DATE = "INDEX_NOT_UP_TO_DATE"
    CHANGE_VECTOR_NOT_MERGED = "CHANGE_VECTOR_NOT_MERGED"
    WAITING_FOR_RESPONSE = "WAITING_FOR_RESPONSE"
    OK = "OK"
    OUT_OF_CPU_CREDITS = "OUT_OF_CPU_CREDITS"
    EARLY_OUT_OF_MEMORY = "EARLY_OUT_OF_MEMORY"
    HIGH_DIRTY_MEMORY = "HIGH_DIRTY_MEMORY"

    def __str__(self):
        return self.value


class ServerOperation(Generic[T]):
    @abstractmethod
    def get_command(self, conventions: "DocumentConventions") -> RavenCommand[T]:
        raise NotImplementedError()


class VoidServerOperation(ServerOperation[None]):
    @abstractmethod
    def get_command(self, conventions: "DocumentConventions") -> "VoidRavenCommand":
        raise NotImplementedError()


class ServerWideOperation(Operation):
    def __init__(
        self, request_executor: "RequestExecutor", conventions: "DocumentConventions", key: int, node_tag: str = None
    ):
        super(ServerWideOperation, self).__init__(request_executor, None, conventions, key, node_tag)
        self.node_tag = node_tag

    def _get_operation_state_command(
        self, conventions: "DocumentConventions", key: int, node_tag: str = None
    ) -> RavenCommand[dict]:
        return GetServerWideOperationStateOperation.GetServerWideOperationStateCommand(key, node_tag)


class GetServerWideOperationStateOperation(ServerOperation[dict]):
    def __init__(self, key: int):
        self.__key = key

    def get_command(self, conventions: "DocumentConventions") -> RavenCommand[dict]:
        return self.GetServerWideOperationStateCommand(self.__key)

    class GetServerWideOperationStateCommand(RavenCommand[dict]):
        def __init__(self, key: int, node_tag: str = None):
            super().__init__(dict)

            self.__key = key
            self._selected_node_tag = node_tag

        def is_read_request(self) -> bool:
            return True

        def create_request(self, node: "ServerNode") -> requests.Request:
            return requests.Request("GET", f"{node.url}/operations/state?id={self.__key}")

        def set_response(self, response: str, from_cache: bool) -> None:
            if response is None:
                return
            self.result = json.loads(response)


class DatabasePutResult:
    def __init__(
        self,
        raft_command_index: Optional[int] = None,
        name: Optional[str] = None,
        topology: Optional[DatabaseTopology] = None,
        nodes_added_to: Optional[List[str]] = None,
    ):
        self.raft_command_index = raft_command_index
        self.name = name
        self.topology = topology
        self.nodes_added_to = nodes_added_to

    @classmethod
    def from_json(cls, json_dict: dict) -> DatabasePutResult:
        return cls(
            json_dict["RaftCommandIndex"],
            json_dict["Name"],
            DatabaseTopology.from_json(json_dict["Topology"]) if "Topology" in json_dict else None,
            json_dict["NodesAddedTo"] if "NodesAddedTo" in json_dict else None,
        )


class CreateDatabaseOperation(ServerOperation):
    def __init__(self, database_record: DatabaseRecord, replication_factor: Optional[int] = 1):
        self.__database_record = database_record
        self.__replication_factor = replication_factor

    def get_command(self, conventions: "DocumentConventions") -> RavenCommand[DatabasePutResult]:
        return CreateDatabaseOperation.CreateDatabaseCommand(
            conventions, self.__database_record, self.__replication_factor
        )

    class CreateDatabaseCommand(RavenCommand[DatabasePutResult], RaftCommand):
        def __init__(
            self,
            conventions: "DocumentConventions",
            database_record: DatabaseRecord,
            replication_factor: int,
            etag: Optional[int] = None,
        ):
            super().__init__(DatabasePutResult)
            self.__conventions = conventions
            self.__database_record = database_record
            self.__replication_factor = replication_factor
            self.__etag = etag
            self.__database_name = database_record.database_name
            if self.__database_name is None:
                raise ValueError("Database name is required.")

        def create_request(self, node: ServerNode) -> requests.Request:
            url = (
                f"{node.url}/admin/databases?name={self.__database_name}"
                f"&replicationFactor={self.__replication_factor}&?raft-request-id={self.get_raft_unique_request_id}"
            )

            request = requests.Request("PUT")
            request.data = self.__database_record.to_json()
            request.headers["Content-type"] = "application/json"
            request.url = url

            if self.__etag is not None:
                request.headers[constants.Headers.ETAG] = f'"{self.__etag}"'

            return request

        def set_response(self, response: str, from_cache: bool) -> None:
            if response is None:
                self._throw_invalid_response()
            self.result = DatabasePutResult.from_json(json.loads(response))

        def is_read_request(self) -> bool:
            return False

        @property
        def get_raft_unique_request_id(self) -> str:
            return RaftIdGenerator.new_id()


class DeleteDatabaseResult:
    def __init__(self, raft_command_index: Optional[int] = None, pending_deletes: Optional[List[str]] = None):
        self.raft_command_index = raft_command_index
        self.pending_deletes = pending_deletes

    @classmethod
    def from_json(cls, json_dict) -> DeleteDatabaseResult:
        return cls(json_dict["RaftCommandIndex"], json_dict["PendingDeletes"])


class DeleteDatabaseOperation(ServerOperation[DeleteDatabaseResult]):
    class Parameters:
        def __init__(
            self,
            database_names: Optional[List[str]] = None,
            hard_delete: Optional[bool] = None,
            from_nodes: Optional[List[str]] = None,
            time_to_wait_for_confirmation: Optional[datetime.timedelta] = None,
        ):
            self.database_names = database_names
            self.hard_delete = hard_delete
            self.from_nodes = from_nodes
            self.time_to_wait_for_confirmation = time_to_wait_for_confirmation

        def to_json(self) -> dict:
            return {
                "DatabaseNames": self.database_names,
                "HardDelete": self.hard_delete,
                "FromNodes": self.from_nodes,
                "TimeToWaitForConfirmation": Utils.timedelta_to_str(self.time_to_wait_for_confirmation),
            }

    def __init__(
        self,
        database_name: Optional[str] = None,
        hard_delete: Optional[bool] = None,
        from_node: Optional[str] = None,
        time_to_wait_for_confirmation: Optional[datetime.timedelta] = None,
        parameters: Optional[DeleteDatabaseOperation.Parameters] = None,
    ):
        if not ((parameters is not None) ^ (database_name is not None)):
            raise ValueError(
                "Please specify either parameters (Parameters object) or database_name (and optionally other args)."
            )
        self.__parameters = (
            parameters
            if parameters
            else DeleteDatabaseOperation.Parameters(
                [database_name], hard_delete, [from_node] if from_node else None, time_to_wait_for_confirmation
            )
        )

    def get_command(self, conventions: "DocumentConventions") -> RavenCommand[DeleteDatabaseResult]:
        return self._DeleteDatabaseCommand(conventions, self.__parameters)

    class _DeleteDatabaseCommand(RavenCommand[DeleteDatabaseResult], RaftCommand):
        def __init__(self, conventions: DocumentConventions, parameters: DeleteDatabaseOperation.Parameters):
            super().__init__(DeleteDatabaseResult)

            if conventions is None:
                raise ValueError("Conventions cannot be None")

            if parameters is None:
                raise ValueError("Parameters cannot be None")

            self.__parameters = parameters

        def is_read_request(self) -> bool:
            return False

        def get_raft_unique_request_id(self) -> str:
            return RaftIdGenerator.new_id()

        def create_request(self, node: ServerNode) -> requests.Request:
            request = requests.Request(
                "DELETE", f"{node.url}/admin/databases?raft-request-id={self.get_raft_unique_request_id()}"
            )
            request.headers["Content-type"] = "application/json"
            request.data = self.__parameters.to_json()
            return request

        def set_response(self, response: str, from_cache: bool) -> None:
            self.result = DeleteDatabaseResult.from_json(json.loads(response))


class BuildNumber:
    def __init__(
        self, product_version: str = None, build_version: int = None, commit_hash: str = None, full_version: str = None
    ):
        self.product_version = product_version
        self.build_version = build_version
        self.commit_hash = commit_hash
        self.full_version = full_version

    @classmethod
    def from_json(cls, json_dict: dict) -> BuildNumber:
        return cls(
            json_dict["ProductVersion"], json_dict["BuildVersion"], json_dict["CommitHash"], json_dict["FullVersion"]
        )


class GetBuildNumberOperation(ServerOperation[BuildNumber]):
    def get_command(self, conventions: "DocumentConventions") -> RavenCommand[BuildNumber]:
        return self.GetBuildNumberCommand()

    class GetBuildNumberCommand(RavenCommand[BuildNumber]):
        def __init__(self):
            super().__init__(BuildNumber)

        def is_read_request(self) -> bool:
            return True

        def create_request(self, node: "ServerNode") -> requests.Request:
            return requests.Request("GET", f"{node.url}/build/version")

        def set_response(self, response: str, from_cache: bool) -> None:
            if response is None:
                self._throw_invalid_response()
            self.result = BuildNumber.from_json(json.loads(response))


class GetDatabaseRecordOperation(ServerOperation[DatabaseRecordWithEtag]):
    def __init__(self, database: str):
        self.__database = database

    def get_command(self, conventions: "DocumentConventions") -> RavenCommand[DatabaseRecordWithEtag]:
        return self.GetDatabaseRecordCommand(self.__database)

    class GetDatabaseRecordCommand(RavenCommand[DatabaseRecordWithEtag]):
        def __init__(self, database_name: str):
            super().__init__(DatabaseRecordWithEtag)
            self.__database_name = database_name

        def is_read_request(self) -> bool:
            return False

        def create_request(self, node: ServerNode) -> requests.Request:
            return requests.Request("GET", f"{node.url}/admin/databases?name={self.__database_name}")

        def set_response(self, response: str, from_cache: bool) -> None:
            if response is None:
                self.result = None
                return
            self.result = DatabaseRecordWithEtag.from_json(json.loads(response))


class GetDatabaseNamesOperation(ServerOperation[List[str]]):
    def __init__(self, start: int, page_size: int):
        super(GetDatabaseNamesOperation, self).__init__()
        self.__start = start
        self.__page_size = page_size

    def get_command(self, conventions: "DocumentConventions") -> RavenCommand[List[str]]:
        return self.__GetDatabaseNamesCommand(self.__start, self.__page_size)

    class __GetDatabaseNamesCommand(RavenCommand[List[str]]):
        def __init__(self, start: int, page_size: int):
            super().__init__(list)
            self.__start = start
            self.__page_size = page_size

        def create_request(self, server_node: ServerNode) -> requests.Request:
            return requests.Request(
                "GET", f"{server_node.url}/databases?start={self.__start}&pageSize={self.__page_size}&namesOnly=true"
            )

        def set_response(self, response: str, from_cache: bool) -> None:
            if response is None:
                raise ValueError("Invalid response")

            response = json.loads(response)

            if "Databases" not in response:
                raise ValueError("Invalid response")

            self.result = response["Databases"]

        def is_read_request(self) -> bool:
            return True


class ModifyOngoingTaskResult:
    def __init__(
        self,
        task_id: Optional[int] = None,
        raft_command_index: Optional[int] = None,
        responsible_node: Optional[str] = None,
    ):
        self.task_id = task_id
        self.raft_command_id = raft_command_index
        self.responsible_node = responsible_node

    @classmethod
    def from_json(cls, json_dict: Dict) -> ModifyOngoingTaskResult:
        return cls(json_dict["TaskId"], json_dict["RaftCommandIndex"], json_dict.get("ResponsibleNode", None))


class DatabaseSettings:
    def __init__(self, settings: Dict[str, str] = None):
        self.settings = settings

    @classmethod
    def from_json(cls, json_dict: Dict) -> DatabaseSettings:
        return cls(json_dict["Settings"])


class ReorderDatabaseMembersOperation(VoidServerOperation):
    class Parameters:
        def __init__(self, members_order: List[str] = None, fixed: bool = None):
            self.members_order = members_order
            self.fixed = fixed

        def to_json(self) -> Dict[str, Any]:
            return {"MembersOrder": self.members_order, "Fixed": self.fixed}

    def __init__(self, database: str = None, order: List[str] = None, fixed: bool = False):
        if not order:
            raise ValueError("Order list must contain values")

        self._database = database
        self._parameters = self.Parameters(order, fixed)

    def get_command(self, conventions: "DocumentConventions") -> "VoidRavenCommand":
        return self.ReorderDatabaseMemberCommand(self._database, self._parameters)

    class ReorderDatabaseMemberCommand(VoidRavenCommand, RaftCommand):
        def __init__(self, database_name: str, parameters: Optional[ReorderDatabaseMembersOperation.Parameters] = None):
            super().__init__()
            if not database_name:
                raise ValueError("Database cannot be empty")

            self._database_name = database_name
            self._parameters = parameters

        def create_request(self, node: ServerNode) -> requests.Request:
            url = f"{node.url}/admin/databases/reorder?name={self._database_name}"
            request = requests.Request("POST", url)
            request.data = self._parameters.to_json()
            return request

        def is_read_request(self) -> bool:
            return False

        def get_raft_unique_request_id(self) -> str:
            return RaftIdGenerator.new_id()


class PromoteDatabaseNodeOperation(ServerOperation[DatabasePutResult]):
    def __init__(self, database_name: str = None, node: str = None):
        self._database_name = database_name
        self._node = node

    def get_command(self, conventions: "DocumentConventions") -> RavenCommand[DatabasePutResult]:
        return self._PromoteDatabaseNodeCommand(self._database_name, self._node)

    class _PromoteDatabaseNodeCommand(RavenCommand[DatabasePutResult], RaftCommand):
        def __init__(self, database_name: str = None, node: str = None):
            super().__init__(DatabasePutResult)
            if not database_name:
                raise ValueError("Database name cannot be None")

            if not node:
                raise ValueError("Node cannot be None")

            self._database_name = database_name
            self._node = node

        def create_request(self, node: ServerNode) -> requests.Request:
            return requests.Request(
                "POST", f"{node.url}/admin/databases/promote?name={self._database_name}&node={self._node}"
            )

        def set_response(self, response: Optional[str], from_cache: bool) -> None:
            if response is None:
                self._throw_invalid_response()

            self.result = DatabasePutResult.from_json(json.loads(response))

        def is_read_request(self) -> bool:
            return False

        def get_raft_unique_request_id(self) -> str:
            return RaftIdGenerator.new_id()


class AddDatabaseNodeOperation(ServerOperation[DatabasePutResult]):
    def __init__(self, database_name: str, node: str = None):
        super().__init__()
        self._database_name = database_name
        self._node = node

    def get_command(self, conventions: "DocumentConventions") -> "AddDatabaseNodeOperation.AddDatabaseNodeCommand":
        return self.AddDatabaseNodeCommand(self._database_name, self._node)

    class AddDatabaseNodeCommand(RavenCommand[DatabasePutResult]):
        def __init__(self, database_name: str, node: Optional[str]):
            super().__init__(DatabasePutResult)

            if not database_name or database_name.isspace():
                raise ValueError("Database name cannot be None or empty")
            self._database_name = database_name
            self._node = node

        def create_request(self, node: ServerNode) -> requests.Request:
            url = f"{node.url}/admin/databases/node?name={self._database_name}"
            if self._node:
                url += f"&node={self._node}"

            return requests.Request("PUT", url)

        def set_response(self, response: Optional[str], from_cache: bool) -> None:
            if response is None:
                self._throw_invalid_response()

            self.result = DatabasePutResult.from_json(json.loads(response))

        def is_read_request(self) -> bool:
            return False

        def get_raft_unique_request_id(self) -> str:
            return RaftIdGenerator.new_id()



from __future__ import annotations
from typing import Optional, Dict, Any


class ExpirationConfiguration:
    def __init__(self, disabled: Optional[bool] = None, delete_frequency_in_sec: Optional[int] = None):
        self.disabled = disabled
        self.delete_frequency_in_sec = delete_frequency_in_sec

    def to_json(self) -> Dict[str, Any]:
        return {"Disabled": self.disabled, "DeleteFrequencyInSec": self.delete_frequency_in_sec}

    @classmethod
    def from_json(cls, json_dict) -> ExpirationConfiguration:
        return cls(json_dict["Disabled"], json_dict["DeleteFrequencyInSec"])



from enum import Enum
from typing import Optional, TYPE_CHECKING

if TYPE_CHECKING:
    from ravendb.http.topology import ClusterTopology


class ServerNode:
    class Role(Enum):
        NONE = "None"
        PROMOTABLE = "Promotable"
        MEMBER = "Member"
        REHAB = "Rehab"

        def __str__(self):
            return self.value

    def __init__(
        self,
        url: str,
        database: Optional[str] = None,
        cluster_tag: Optional[str] = None,
        server_role: Optional[Role] = None,
    ):
        self.url = url
        self.database = database
        self.cluster_tag = cluster_tag
        self.server_role = server_role
        self.__last_server_version_check = 0
        self.__last_server_version: str = None

    def __eq__(self, other) -> bool:
        if self == other:
            return True
        if other is None or type(self) != type(other):
            return False
        if self.url != other.url if self.url is not None else other.url is not None:
            return False
        return self.database == other.database if self.database is not None else other.database is None

    def __hash__(self) -> int:
        result = self.url.__hash__() if self.url else 0
        result = 31 * result + self.database.__hash__() if self.database is not None else 0
        return result

    @property
    def last_server_version(self) -> str:
        return self.__last_server_version

    @classmethod
    def create_from(cls, topology: "ClusterTopology"):
        nodes = []
        if topology is None:
            return nodes

        for key, value in topology.members.items():
            server_node = cls(value, cluster_tag=key)
            nodes.append(server_node)

        for key, value in topology.watchers.items():
            server_node = cls(value, cluster_tag=key)
            nodes.append(server_node)

        return nodes

    def should_update_server_version(self) -> bool:
        if self.last_server_version is None or self.__last_server_version_check > 100:
            return True

        self.__last_server_version_check += 1
        return False

    def update_server_version(self, server_version: str):
        self.__last_server_version = server_version
        self.__last_server_version_check = 0

    def discard_server_version(self) -> None:
        self.__last_server_version_check = None
        self.__last_server_version_check = 0



from __future__ import annotations
from enum import Enum
from typing import Optional, Union

from ravendb.http.misc import ReadBalanceBehavior, LoadBalanceBehavior


class StudioEnvironment(Enum):
    NONE = "NONE"
    DEVELOPMENT = "DEVELOPMENT"
    TESTING = "TESTING"
    PRODUCTION = "PRODUCTION"


class StudioConfiguration:
    def __init__(self, disabled: Optional[bool] = None, environment: Optional[StudioEnvironment] = None):
        self.disabled = disabled
        self.environment = environment


class ClientConfiguration:
    def __init__(self):
        self.__identity_parts_separator: Union[None, str] = None
        self.etag: int = 0
        self.disabled: bool = False
        self.max_number_of_requests_per_session: Optional[int] = None
        self.read_balance_behavior: Optional[ReadBalanceBehavior] = None
        self.load_balance_behavior: Optional[LoadBalanceBehavior] = None
        self.load_balancer_context_seed: Optional[int] = None

    @property
    def identity_parts_separator(self) -> str:
        return self.__identity_parts_separator

    @identity_parts_separator.setter
    def identity_parts_separator(self, value: str):
        if value is not None and "|" == value:
            raise ValueError("Cannot set identity parts separator to '|'")
        self.__identity_parts_separator = value

    def to_json(self) -> dict:
        return {
            "IdentityPartsSeparator": self.__identity_parts_separator,
            "Etag": self.etag,
            "Disabled": self.disabled,
            "MaxNumberOfRequestsPerSession": self.max_number_of_requests_per_session,
            "ReadBalanceBehavior": (
                self.read_balance_behavior.value if self.read_balance_behavior else ReadBalanceBehavior.NONE
            ),
            "LoadBalanceBehavior": (
                self.load_balance_behavior.value if self.load_balance_behavior else LoadBalanceBehavior.NONE
            ),
            "LoadBalancerContextSeed": self.load_balancer_context_seed,
        }

    @classmethod
    def from_json(cls, json_dict: dict) -> Optional[ClientConfiguration]:
        if json_dict is None:
            return None
        config = cls()
        config.__identity_parts_separator = json_dict["IdentityPartsSeparator"]
        config.etag = json_dict["Etag"]
        config.disabled = json_dict["Disabled"]
        config.max_number_of_requests_per_session = json_dict["MaxNumberOfRequestsPerSession"]
        config.read_balance_behavior = ReadBalanceBehavior(json_dict["ReadBalanceBehavior"])
        config.load_balance_behavior = LoadBalanceBehavior(json_dict["LoadBalanceBehavior"])
        config.load_balancer_context_seed = json_dict["LoadBalancerContextSeed"]

        return config



from __future__ import annotations
import json
from typing import List, TYPE_CHECKING, Optional, Tuple

import requests

from ravendb.exceptions import exceptions
from ravendb.exceptions.exceptions import ErrorResponseException
from ravendb.documents.indexes.definitions import (
    IndexDefinition,
    IndexErrors,
    IndexLockMode,
    IndexPriority,
    IndexRunningStatus,
)
from ravendb.documents.operations.definitions import MaintenanceOperation, VoidMaintenanceOperation
from ravendb.http.raven_command import RavenCommand, VoidRavenCommand
from ravendb.http.server_node import ServerNode
from ravendb.http.topology import RaftCommand
from ravendb.tools.utils import Utils
from ravendb.util.util import RaftIdGenerator
from ravendb.documents.indexes.stats import IndexStats

if TYPE_CHECKING:
    from ravendb.documents.conventions import DocumentConventions


class PutIndexesOperation(MaintenanceOperation):
    def __init__(self, *indexes_to_add: IndexDefinition):
        if len(indexes_to_add) == 0:
            raise ValueError("Invalid indexes_to_add")

        super(PutIndexesOperation, self).__init__()
        self._indexes_to_add = indexes_to_add
        self.__all_java_script_indexes = True  # todo: set it in the command

    def get_command(self, conventions: "DocumentConventions") -> RavenCommand[List]:
        return self.__PutIndexesCommand(conventions, self._indexes_to_add)

    class __PutIndexesCommand(RavenCommand[List], RaftCommand):
        def __init__(self, conventions: "DocumentConventions", index_to_add: Tuple[IndexDefinition]):
            super().__init__(list)
            if conventions is None:
                raise ValueError("Conventions cannot be None")
            if index_to_add is None:
                raise ValueError("None indexes_to_add is not valid")

            self.indexes_to_add = []
            for index_definition in index_to_add:
                if not isinstance(index_definition, IndexDefinition):
                    raise ValueError("index_definition in indexes_to_add must be IndexDefinition type")
                if index_definition.name is None:
                    raise ValueError("None Index name is not valid")
                self.indexes_to_add.append(index_definition.to_json())

        def get_raft_unique_request_id(self) -> str:
            return RaftIdGenerator.new_id()

        def is_read_request(self) -> bool:
            return False

        def create_request(self, node: ServerNode) -> requests.Request:
            request = requests.Request("PUT")
            request.url = f"{node.url}/databases/{node.database}/admin/indexes"
            request.data = {"Indexes": self.indexes_to_add}
            request.headers["Content-type"] = "application/json"
            return request

        def set_response(self, response: str, from_cache: bool) -> None:
            self.result = json.loads(response)  # todo: PutIndexResult instead of dict
            if "Error" in response:
                raise ErrorResponseException(response["Error"])


class GetIndexNamesOperation(MaintenanceOperation):
    def __init__(self, start, page_size):
        super().__init__()
        self._start = start
        self._page_size = page_size

    def get_command(self, conventions) -> RavenCommand[List[str]]:
        return self.__GetIndexNamesCommand(self._start, self._page_size)

    class __GetIndexNamesCommand(RavenCommand[List[str]]):
        def __init__(self, start, page_size):
            super().__init__(list)
            self._start = start
            self._page_size = page_size

        def create_request(self, server_node):
            return requests.Request(
                "GET",
                f"{server_node.url}/databases/{server_node.database}/indexes?start={self._start}&pageSize={self._page_size}&namesOnly=true",
            )

        def is_read_request(self) -> bool:
            return True

        def set_response(self, response: str, from_cache: bool) -> None:
            if response is None:
                raise ValueError("Invalid response")
            response = json.loads(response)
            if "Error" in response:
                raise exceptions.ErrorResponseException(response["Error"])
            if "Results" not in response:
                raise ValueError("Invalid response")
            self.result = response["Results"]


class DeleteIndexOperation(VoidMaintenanceOperation):
    def __init__(self, index_name: str):
        if index_name is None:
            raise ValueError("Index name cannot be None")
        super(DeleteIndexOperation, self).__init__()
        self.__index_name = index_name

    def get_command(self, conventions) -> VoidRavenCommand:
        return self.__DeleteIndexCommand(self.__index_name)

    class __DeleteIndexCommand(VoidRavenCommand, RaftCommand):
        def __init__(self, index_name: str):
            if not index_name:
                raise ValueError("Invalid index_name")
            super().__init__()
            self.__index_name = index_name

        def create_request(self, server_node: ServerNode) -> requests.Request:
            return requests.Request(
                "DELETE",
                f"{server_node.url}/databases/{server_node.database}/indexes?name={Utils.quote_key(self.__index_name)}",
            )

        def get_raft_unique_request_id(self) -> str:
            return RaftIdGenerator.new_id()


class DisableIndexOperation(VoidMaintenanceOperation):
    def __init__(self, index_name: str, cluster_wide: bool = False):
        if index_name is None:
            raise ValueError("index_name cannot be None")
        super().__init__()
        self.__index_name = index_name
        self.__cluster_wide = cluster_wide

    def get_command(self, conventions) -> VoidRavenCommand:
        return self.__DisableIndexCommand(self.__index_name, self.__cluster_wide)

    class __DisableIndexCommand(VoidRavenCommand, RaftCommand):
        def __init__(self, index_name: str, cluster_wide: bool):
            if index_name is None:
                raise ValueError("index_name cannot be None")
            super().__init__()
            self.__index_name = index_name
            self.__cluster_wide = cluster_wide

        def create_request(self, server_node) -> requests.Request:
            return requests.Request(
                "POST",
                f"{server_node.url}/databases/{server_node.database}/admin/indexes"
                f"/disable?name={Utils.escape(self.__index_name, False, False)}"
                f"&clusterWide={self.__cluster_wide}",
            )

        def get_raft_unique_request_id(self) -> str:
            return RaftIdGenerator.new_id()


class EnableIndexOperation(VoidMaintenanceOperation):
    def __init__(self, index_name: str, cluster_wide: bool = False):
        if index_name is None:
            raise ValueError("index_name cannot be None")
        super().__init__()
        self.__index_name = index_name
        self.__cluster_wide = cluster_wide

    def get_command(self, conventions) -> VoidRavenCommand:
        return self.__EnableIndexCommand(self.__index_name, self.__cluster_wide)

    class __EnableIndexCommand(VoidRavenCommand):
        def __init__(self, index_name: str, cluster_wide: bool):
            if index_name is None:
                raise ValueError("index_name cannot be None")
            super().__init__()
            self.__index_name = index_name
            self.__cluster_wide = cluster_wide

        def create_request(self, server_node) -> requests.Request:
            return requests.Request(
                "POST",
                f"{server_node.url}/databases/{server_node.database}/admin/indexes"
                f"/enable?name={Utils.escape(self.__index_name, False, False)}"
                f"&clusterWide={self.__cluster_wide}",
            )

        def raft_unique_request_id(self) -> str:
            return RaftIdGenerator.new_id()


class StopIndexingOperation(VoidMaintenanceOperation):
    def __init__(self):
        super().__init__()

    def get_command(self, conventions) -> VoidRavenCommand:
        return self.__StopIndexingCommand()

    class __StopIndexingCommand(VoidRavenCommand):
        def __init__(self):
            super().__init__()

        def create_request(self, server_node) -> requests.Request:
            return requests.Request("POST", f"{server_node.url}/databases/{server_node.database}/admin/indexes/stop")


class StartIndexingOperation(VoidMaintenanceOperation):
    def __init__(self):
        super().__init__()

    def get_command(self, conventions) -> VoidRavenCommand:
        return self.__StopIndexingCommand()

    class __StopIndexingCommand(VoidRavenCommand):
        def __init__(self):
            super().__init__()

        def create_request(self, server_node) -> requests.Request:
            return requests.Request("POST", f"{server_node.url}/databases/{server_node.database}/admin/indexes/start")


class StartIndexOperation(VoidMaintenanceOperation):
    def __init__(self, index_name: str):
        if not index_name:
            raise ValueError("Index name can't be None or empty")
        super().__init__()
        self.__index_name = index_name

    def get_command(self, conventions) -> VoidRavenCommand:
        return self.__StartIndexCommand(self.__index_name)

    class __StartIndexCommand(VoidRavenCommand):
        def __init__(self, index_name):
            super().__init__()
            if not index_name:
                raise ValueError("Index name can't be None or empty")
            self._index_name = index_name

        def create_request(self, server_node) -> requests.Request:
            return requests.Request(
                "POST",
                f"{server_node.url}/databases/{server_node.database}"
                f"/admin/indexes/start?name={Utils.quote_key(self._index_name)}",
            )


class StopIndexOperation(VoidMaintenanceOperation):
    def __init__(self, index_name: str):
        if not index_name:
            raise ValueError("Index name can't be None or empty")
        super().__init__()
        self.__index_name = index_name

    def get_command(self, conventions) -> VoidRavenCommand:
        return self.__StopIndexCommand(self.__index_name)

    class __StopIndexCommand(VoidRavenCommand):
        def __init__(self, index_name):
            super().__init__()
            if not index_name:
                raise ValueError("Index name can't be None or empty")
            self._index_name = index_name

        def create_request(self, server_node) -> requests.Request:
            return requests.Request(
                "POST",
                f"{server_node.url}/databases/{server_node.database}"
                f"/admin/indexes/stop?name={Utils.quote_key(self._index_name)}",
            )


class IndexStatus:
    def __init__(self, name: str = None, status: IndexRunningStatus = None):
        self.name = name
        self.status = status

    @classmethod
    def from_json(cls, json_dict: dict) -> IndexStatus:
        return cls(json_dict["Name"], IndexRunningStatus(json_dict["Status"]))


class IndexingStatus:
    def __init__(self, status: IndexRunningStatus = None, indexes: List[IndexStatus] = None):
        self.status = status
        self.indexes = indexes

    @classmethod
    def from_json(cls, json_dict) -> IndexingStatus:
        return cls(IndexRunningStatus(json_dict["Status"]), list(map(IndexStatus.from_json, json_dict["Indexes"])))


class GetIndexingStatusOperation(MaintenanceOperation[IndexingStatus]):
    def __init__(self):
        super(GetIndexingStatusOperation, self).__init__()

    def get_command(self, conventions: "DocumentConventions") -> RavenCommand[IndexingStatus]:
        return self.__GetIndexingStatusCommand()

    class __GetIndexingStatusCommand(RavenCommand[IndexingStatus]):
        def __init__(self):
            super().__init__(IndexingStatus)

        def create_request(self, server_node: ServerNode) -> requests.Request:
            return requests.Request("GET", f"{server_node.url}/databases/{server_node.database}/indexes/status")

        def is_read_request(self) -> bool:
            pass

        def set_response(self, response: str, from_cache: bool) -> None:
            if response is None:
                raise ValueError("Invalid response")
            response = json.loads(response)
            if "Error" in response:
                raise exceptions.ErrorResponseException(response["Error"])
            self.result = IndexingStatus.from_json(response)


class GetIndexStatisticsOperation(MaintenanceOperation["IndexStats"]):
    def __init__(self, name):
        if name is None:
            raise ValueError("Index name cannot be None")
        super().__init__()
        self.__index_name = name

    def get_command(self, conventions) -> RavenCommand["IndexStats"]:
        return self._GetIndexStatisticsCommand(self.__index_name)

    class _GetIndexStatisticsCommand(RavenCommand["IndexStats"]):
        def __init__(self, index_name: str):
            if not index_name:
                raise ValueError("Index name can't be None or empty")
            super().__init__(dict)
            self.__index_name = index_name

        def create_request(self, server_node) -> requests.Request:
            url = f"{server_node.url}/databases/{server_node.database}/indexes/stats?name={Utils.quote_key(self.__index_name)}"
            return requests.Request(
                "GET",
                url,
            )

        def is_read_request(self) -> bool:
            return True

        def set_response(self, response: str, from_cache: bool) -> None:
            if response is None:
                self._throw_invalid_response()

            response = json.loads(response)
            results = response["Results"]

            if len(results) != 1:
                self._throw_invalid_response()

            self.result = IndexStats.from_json(results[0])


class GetIndexesStatisticsOperation(
    MaintenanceOperation[dict]
):  # replace returned type with IndexStats -> its enormous
    def get_command(self, conventions) -> RavenCommand[dict]:
        return self.__GetIndexesStatisticsCommand()

    class __GetIndexesStatisticsCommand(RavenCommand[dict]):
        def __init__(self):
            super().__init__(dict)

        def create_request(self, server_node) -> requests.Request:
            return requests.Request("GET", f"{server_node.url}/databases/{server_node.database}/indexes/stats")

        def is_read_request(self) -> bool:
            return True

        def set_response(self, response: str, from_cache: bool) -> None:
            if response is None:
                raise ValueError("Invalid response")

            response = json.loads(response)
            if "Error" in response:
                raise exceptions.ErrorResponseException(response["Error"])
            self.result = response["Results"]


class GetIndexErrorsOperation(MaintenanceOperation[List[IndexErrors]]):
    def __init__(self, *index_names: str):
        super(GetIndexErrorsOperation, self).__init__()
        self._index_names = index_names

    def get_command(self, conventions: "DocumentConventions") -> RavenCommand[List[IndexErrors]]:
        return self.__GetIndexErrorsCommand(*self._index_names)

    class __GetIndexErrorsCommand(RavenCommand[List[IndexErrors]]):
        def __init__(self, *index_names: str):
            super().__init__(list)
            self._index_names = index_names

        def create_request(self, server_node) -> requests.Request:
            url = f"{server_node.url}/databases/{server_node.database}/indexes/errors"
            if self._index_names and len(self._index_names) != 0:
                url += "?"
                for index_name in self._index_names:
                    url += f"&name={Utils.escape(index_name, None, None)}"
            return requests.Request("GET", url)

        def set_response(self, response: str, from_cache: bool):
            if response is None:
                raise ValueError("InvalidResponse")
            results = json.loads(response)["Results"]
            self.result = []
            for result in results:
                self.result.append(IndexErrors.from_json(result))

        def is_read_request(self) -> bool:
            return True


class GetIndexesOperation(MaintenanceOperation[List[IndexDefinition]]):
    def __init__(self, start: int, page_size: int):
        super(GetIndexesOperation, self).__init__()
        self.__start = start
        self.__page_size = page_size

    def get_command(self, conventions):
        return self.__GetIndexesCommand(self.__start, self.__page_size)

    class __GetIndexesCommand(RavenCommand[List[IndexDefinition]]):
        def __init__(self, start: int, page_size: int):
            super().__init__(list)
            self.__start = start
            self.__page_size = page_size

        def create_request(self, server_node) -> requests.Request:
            return requests.Request(
                "GET",
                f"{server_node.url}/databases/{server_node.database}"
                f"/indexes?start={self.__start}&pageSize{self.__page_size} ",
            )

        def set_response(self, response: str, from_cache: bool) -> None:
            response = json.loads(response)["Results"]
            result = []
            for index in response:
                result.append(IndexDefinition.from_json(index))
            self.result = result

        def is_read_request(self) -> bool:
            return True


class GetIndexOperation(MaintenanceOperation[IndexDefinition]):
    def __init__(self, index_name: str):
        if not index_name:
            raise ValueError("Index name can't be None or empty")
        super().__init__()
        self.__index_name = index_name

    def get_command(self, conventions: "DocumentConventions"):
        return self.__GetIndexCommand(self.__index_name)

    class __GetIndexCommand(RavenCommand[IndexDefinition]):
        def __init__(self, index_name: str):
            if not index_name:
                raise ValueError("Index name can't be None or empty")
            super().__init__(IndexDefinition)
            self.__index_name = index_name

        def create_request(self, server_node) -> requests.Request:
            return requests.Request(
                "GET",
                f"{server_node.url}/databases/{server_node.database}"
                f"/indexes?name={Utils.escape(self.__index_name,False,False)}",
            )

        def set_response(self, response: str, from_cache: bool) -> None:
            if not response:
                return
            self.result = IndexDefinition.from_json(json.loads(response)["Results"][0])

        def is_read_request(self) -> bool:
            return True


class SetIndexesLockOperation(VoidMaintenanceOperation):
    def __init__(self, mode: IndexLockMode, *index_names: str):
        if len(index_names) == 0:
            raise ValueError("Index names count cannot be zero. Pass at least one index name.")
        super(SetIndexesLockOperation, self).__init__()
        self.__mode = mode
        self.__index_names = index_names
        self.__filter_auto_indexes()

    def get_command(self, conventions: "DocumentConventions") -> VoidRavenCommand:
        return self.__SetIndexesLockCommand(self.__mode, *self.__index_names)

    def __filter_auto_indexes(self):
        for name in self.__index_names:
            if name.startswith("auto/"):
                raise ValueError("Index list contains Auto-Indexes. Lock Mode is not set for Auto-Indexes")

    class __SetIndexesLockCommand(VoidRavenCommand, RaftCommand):
        def __init__(self, mode: IndexLockMode, *index_names: str):
            super().__init__()
            self._mode = mode
            self._index_names = index_names

        def is_read_request(self) -> bool:
            return False

        def get_raft_unique_request_id(self) -> str:
            return RaftIdGenerator.new_id()

        def create_request(self, server_node):
            request = requests.Request("POST")
            request.url = f"{server_node.url}/databases/{server_node.database}/indexes/set-lock"
            request.data = {"IndexNames": self._index_names, "Mode": self._mode}
            return request


class SetIndexesPriorityOperation(VoidMaintenanceOperation):
    def __init__(self, priority: IndexPriority, *index_names: str):
        if len(index_names) == 0:
            raise ValueError("Index names count cannot be zero. Pass at least one index name.")
        super(SetIndexesPriorityOperation, self).__init__()
        self._priority = priority
        self._index_names = index_names

    def get_command(self, conventions):
        return self.__SetIndexesPriorityCommand(self._priority, *self._index_names)

    class __SetIndexesPriorityCommand(VoidRavenCommand, RaftCommand):
        def __init__(self, priority: IndexPriority, *index_names: str):
            super().__init__()
            self._priority = priority
            self._index_names = index_names

        def create_request(self, server_node):
            request = requests.Request("POST")
            request.url = f"{server_node.url}/databases/{server_node.database}/indexes/set-priority"
            request.data = {"IndexNames": self._index_names, "Priority": self._priority}
            return request

        def is_read_request(self) -> bool:
            return False

        def get_raft_unique_request_id(self) -> str:
            return RaftIdGenerator.new_id()


class GetTermsOperation(MaintenanceOperation[List[str]]):
    def __init__(self, index_name: str, field: str, from_value: Optional[str], page_size: int = None):
        if index_name is None:
            raise ValueError("Index name cannot be None")
        if field is None:
            raise ValueError("Field cannot be null")
        super(GetTermsOperation, self).__init__()
        self.__index_name = index_name
        self.__field = field
        self.__from_value = from_value
        self.__page_size = page_size

    def get_command(self, conventions):
        return self.__GetTermsCommand(self.__index_name, self.__field, self.__from_value, self.__page_size)

    class __GetTermsCommand(RavenCommand[List[str]]):
        def __init__(self, index_name: str, field: str, from_value: str, page_size: int):
            super().__init__(list)
            self.__index_name = index_name
            self.__field = field
            self.__from_value = from_value
            self.__page_size = page_size

        def create_request(self, server_node) -> requests.Request:
            return requests.Request(
                "GET",
                f"{server_node.url}/databases/{server_node.database}"
                f"/indexes/terms?name={Utils.escape(self.__index_name, False, False)}"
                f"&field={Utils.escape(self.__field, False, False)}"
                f"&fromValue={self.__from_value or ''}"
                f"&pageSize={self.__page_size or ''}",
            )

        def set_response(self, response: str, from_cache: bool) -> None:
            if response is None:
                raise ValueError("Invalid response")
            response = json.loads(response)
            self.result = response["Terms"]


class IndexHasChangedOperation(MaintenanceOperation[bool]):
    def __init__(self, index: IndexDefinition):
        if index is None:
            raise ValueError("Index cannot be null")
        super(IndexHasChangedOperation, self).__init__()
        self.__index = index

    def get_command(self, conventions: "DocumentConventions") -> RavenCommand[bool]:
        return self.__IndexHasChangedCommand(self.__index)

    class __IndexHasChangedCommand(RavenCommand[bool]):
        def __init__(self, index: IndexDefinition):
            super().__init__(bool)
            self.__index = index

        def is_read_request(self) -> bool:
            return False

        def create_request(self, server_node: ServerNode) -> requests.Request:
            request = requests.Request("POST")
            request.url = f"{server_node.url}/databases/{server_node.database}/indexes/has-changed"
            request.data = {
                "Configuration": self.__index.configuration,
                "Fields": self.__index.fields,
                "LockMode": self.__index.lock_mode,
                "Maps": self.__index.maps,
                "Name": self.__index.name,
                "OutputReduceToCollection": self.__index.output_reduce_to_collection,
                "Priority": self.__index.priority,
                "Reduce": self.__index.reduce,
                "SourceType": self.__index.source_type,
                "Type": self.__index.type,
            }
            return request

        def set_response(self, response: str, from_cache: bool) -> None:
            if not response:
                raise ValueError("Response is invalid")

            self.result = json.loads(response)["Changed"]


class ResetIndexOperation(VoidMaintenanceOperation):
    def __init__(self, index_name: str):
        if index_name is None:
            raise ValueError("Index name cannot be None")

        self._index_name = index_name

    def get_command(self, conventions: "DocumentConventions") -> "VoidRavenCommand":
        return ResetIndexOperation.ResetIndexCommand(self._index_name)

    class ResetIndexCommand(VoidRavenCommand):
        def __init__(self, index_name: str):
            super().__init__()
            if index_name is None:
                raise ValueError("Index name cannot be None")

            self._index_name = index_name

        def create_request(self, node: ServerNode) -> requests.Request:
            return requests.Request("RESET", f"{node.url}/databases/{node.database}/indexes?name={self._index_name}")


class DeleteIndexErrorsOperation(VoidMaintenanceOperation):
    def __init__(self, index_names: List[str] = None):
        self._index_names = index_names

    def get_command(self, conventions: "DocumentConventions") -> "VoidRavenCommand":
        return self.DeleteIndexErrorsCommand(self._index_names)

    class DeleteIndexErrorsCommand(VoidRavenCommand):
        def __init__(self, index_names: List[str]):
            super().__init__()
            self._index_names = index_names

        def create_request(self, node: ServerNode) -> requests.Request:
            url = f"{node.url}/databases/{node.database}/indexes/errors"

            if self._index_names is not None and len(self._index_names) > 0:
                url = f"{url}?"
                for index_name in self._index_names:
                    url = f"{url}&name={Utils.quote_key(index_name)}"

            return requests.Request("DELETE", url)

        def is_read_request(self) -> bool:
            return False



from __future__ import annotations

import json
from typing import Any, Dict, TYPE_CHECKING, Optional

import requests

from ravendb import DocumentsCompressionConfiguration, RaftCommand, ServerNode
from ravendb.documents.operations.definitions import MaintenanceOperation
from ravendb.util.util import RaftIdGenerator
from ravendb.http.raven_command import RavenCommand

if TYPE_CHECKING:
    from ravendb.documents.conventions import DocumentConventions


class DocumentCompressionConfigurationResult:
    def __init__(self, raft_command_index: int = None):
        self.raft_command_index = raft_command_index

    @classmethod
    def from_json(cls, json_dict: Dict[str, Any]) -> DocumentCompressionConfigurationResult:
        return cls(json_dict["RaftCommandIndex"])


class UpdateDocumentsCompressionConfigurationOperation(MaintenanceOperation[DocumentCompressionConfigurationResult]):
    def __init__(self, configuration: DocumentsCompressionConfiguration):
        if configuration is None:
            raise ValueError("Configuration cannot be None")
        self._documents_compression_configuration = configuration

    def get_command(self, conventions: "DocumentConventions") -> "RavenCommand[DocumentCompressionConfigurationResult]":
        return self.UpdateDocumentCompressionConfigurationCommand(self._documents_compression_configuration)

    class UpdateDocumentCompressionConfigurationCommand(
        RavenCommand[DocumentCompressionConfigurationResult], RaftCommand
    ):
        def __init__(self, configuration: DocumentsCompressionConfiguration):
            super().__init__(DocumentCompressionConfigurationResult)

            if configuration is None:
                raise ValueError("Configuration cannot be None")

            self._documents_compression_configuration = configuration

        def is_read_request(self) -> bool:
            return False

        def create_request(self, node: ServerNode) -> requests.Request:
            url = f"{node.url}/databases/{node.database}/admin/documents-compression/config"

            request = requests.Request("POST", url)

            request.data = self._documents_compression_configuration.to_json()

            return request

        def set_response(self, response: Optional[str], from_cache: bool) -> None:
            if response is None:
                self._throw_invalid_response()

            self.result = DocumentCompressionConfigurationResult.from_json(json.loads(response))

        def get_raft_unique_request_id(self) -> str:
            return RaftIdGenerator.new_id()



from __future__ import annotations

import json
from typing import TYPE_CHECKING

import requests

from ravendb.documents.operations.configuration.definitions import ClientConfiguration
from ravendb.documents.operations.definitions import VoidMaintenanceOperation, MaintenanceOperation
from ravendb.http.raven_command import RavenCommand, VoidRavenCommand
from ravendb.http.server_node import ServerNode
from ravendb.http.topology import RaftCommand
from ravendb.util.util import RaftIdGenerator

if TYPE_CHECKING:
    from ravendb.documents.conventions import DocumentConventions

from ravendb.serverwide.operations.common import ServerOperation, VoidServerOperation


class GetClientConfigurationOperation(MaintenanceOperation):
    def get_command(self, conventions: "DocumentConventions") -> "RavenCommand[GetClientConfigurationOperation.Result]":
        return self.GetClientConfigurationCommand()

    class GetClientConfigurationCommand(RavenCommand):
        def __init__(self):
            super().__init__(GetClientConfigurationOperation.Result)

        def is_read_request(self) -> bool:
            return False

        def create_request(self, node: ServerNode) -> requests.Request:
            return requests.Request(method="GET", url=f"{node.url}/databases/{node.database}/configuration/client")

        def set_response(self, response: str, from_cache: bool) -> None:
            if response is None:
                return
            self.result = GetClientConfigurationOperation.Result.from_json(json.loads(response))

    class Result:
        def __init__(self, etag: int, configuration: ClientConfiguration):
            self.etag = etag
            self.configuration = configuration

        @classmethod
        def from_json(cls, json_dict: dict) -> GetClientConfigurationOperation.Result:
            return cls(json_dict["Etag"], ClientConfiguration.from_json(json_dict["Configuration"]))


class PutClientConfigurationOperation(VoidMaintenanceOperation):
    def __init__(self, config: ClientConfiguration):
        if config is None:
            raise ValueError("Configuration cannot be None")
        self.__configuration = config

    def get_command(self, conventions: "DocumentConventions") -> "VoidRavenCommand":
        return self.__PutClientConfigurationCommand(conventions, self.__configuration)

    class __PutClientConfigurationCommand(VoidRavenCommand, RaftCommand):
        def __init__(self, conventions: "DocumentConventions", config: ClientConfiguration):
            super().__init__()
            if conventions is None:
                raise ValueError("Conventions cannot be None")

            if config is None:
                raise ValueError("Configuration cannot be None")

            self.__configuration = json.dumps(config.to_json(), default=conventions.json_default_method)

        def create_request(self, node: ServerNode) -> requests.Request:
            return requests.Request(
                "PUT", f"{node.url}/databases/{node.database}/admin/configuration/client", data=self.__configuration
            )

        def get_raft_unique_request_id(self) -> str:
            return RaftIdGenerator.new_id()


class PutServerWideClientConfigurationOperation(VoidServerOperation):
    def __init__(self, config: ClientConfiguration):
        if config is None:
            raise ValueError("Configuration cannot be None")
        self.__configuration = config

    def get_command(self, conventions: "DocumentConventions") -> "VoidRavenCommand":
        return self.__PutServerWideClientConfigurationCommand(conventions, self.__configuration)

    class __PutServerWideClientConfigurationCommand(VoidRavenCommand, RaftCommand):
        def __init__(self, conventions: "DocumentConventions", config: ClientConfiguration):
            super().__init__()
            if conventions is None:
                raise ValueError("Conventions cannot be None")

            if config is None:
                raise ValueError("Configuration cannot be None")

            self.__configuration = json.dumps(config.to_json(), default=conventions.json_default_method)

        def create_request(self, node: ServerNode) -> requests.Request:
            return requests.Request("PUT", f"{node.url}/admin/configuration/client", data=self.__configuration)

        def get_raft_unique_request_id(self) -> str:
            return RaftIdGenerator.new_id()


class GetServerWideClientConfigurationOperation(ServerOperation[ClientConfiguration]):
    def get_command(self, conventions: "DocumentConventions") -> RavenCommand[ClientConfiguration]:
        return GetServerWideClientConfigurationOperation.__GetServerWideClientConfigurationCommand()

    class __GetServerWideClientConfigurationCommand(RavenCommand[ClientConfiguration]):
        def __init__(self):
            super().__init__(ClientConfiguration)

        def is_read_request(self) -> bool:
            return False

        def create_request(self, node: ServerNode) -> requests.Request:
            return requests.Request(method="GET", url=f"{node.url}/configuration/client")

        def set_response(self, response: str, from_cache: bool) -> None:
            if response is None:
                return
            self.result = ClientConfiguration.from_json(json.loads(response))



import json
from typing import TYPE_CHECKING

import requests

from ravendb.documents.queries.query import QueryResult
from ravendb.extensions.json_extensions import JsonExtensions
from ravendb.http.raven_command import RavenCommand
from ravendb.http.server_node import ServerNode

if TYPE_CHECKING:
    from ravendb.documents.session.document_session import InMemoryDocumentSessionOperations
    from ravendb.documents.queries.index_query import IndexQuery


class QueryCommand(RavenCommand[QueryResult]):
    def __init__(
        self,
        session: "InMemoryDocumentSessionOperations",
        index_query: "IndexQuery",
        metadata_only: bool,
        index_entries_only: bool,
    ):
        super().__init__(QueryResult)
        self.__session = session
        if index_query is None:
            raise ValueError("index_query cannot be None")

        self.__index_query = index_query
        self.__metadata_only = metadata_only
        self.__index_entries_only = index_entries_only

    def create_request(self, node: ServerNode) -> requests.Request:
        self._can_cache = not self.__index_query.disable_caching
        self._can_cache_aggressively = self._can_cache and not self.__index_query.wait_for_non_stale_results

        path = [node.url, "/databases/", node.database, "/queries?queryHash=", self.__index_query.get_query_hash()]

        if self.__metadata_only:
            path.append("&metadataOnly=true")

        if self.__index_entries_only:
            path.append("&debug=entries")

        request = requests.Request("POST", "".join(path))
        request.data = JsonExtensions.write_index_query(self.__session.conventions, self.__index_query)
        return request

    def set_response(self, response: str, from_cache: bool) -> None:
        if response is None:
            self.result = None
            return

        self.result = QueryResult.from_json(json.loads(response))
        if from_cache:
            self.result.duration_in_ms = -1

            if self.result.timings is not None:
                self.result.timings.duration_in_ms = -1
                self.result.timings = None

    def is_read_request(self) -> bool:
        return True
